import os
import string
import hashlib
from os import stat_result

import gnupg

checksum_prefix_extension = '.checksum'


def create_checksum_file(file_name_with_path: str, pgp_key_id: str = None) -> str:
    """
    Creates a checksum file for a given file. The file can be either signed or unsigned,
    depending on whether a PGP key ID is provided. If no PGP key ID is provided, an
    unsigned checksum file is created. If a PGP key ID is specified, the checksum file
    is signed using the provided PGP key.

    :param file_name_with_path: The full file path of the file for which the checksum will
        be generated. this can be an implicit path by just naming the file on which this needs to be applied.
    :type file_name_with_path: str
    :param pgp_key_id: The PGP key ID for signing the checksum file. If None, the checksum
        will be unsigned.
    :type pgp_key_id: str, optional
    :return: The name of the created checksum file without containing the full file path.
    :rtype: str
    """
    file_name_without_extension = os.path.splitext(os.path.basename(file_name_with_path))[0]
    checksum_file_name = file_name_without_extension + checksum_prefix_extension
    if pgp_key_id is None:
        __create_named_unsigned_checksum_file(file_name_with_path, checksum_file_name)
    else:
        __create_named_signed_checksum_file(file_name_with_path, checksum_file_name, pgp_key_id)
    return checksum_file_name


def __create_named_signed_checksum_file(file_name: str, checksum_file_name: str, pgp_key_id: str):
    """
    Creates a named and signed checksum file for a given file. This function generates a checksum
    file with the specified name, signs it using the provided PGP key ID, and ensures the file
    is securely prepared for validation or distribution. The process involves creating an unsigned
    checksum file first and then utilizing the signing method for integrity.

    :param file_name: Name of the file for which the checksum is to be created.
    :type file_name: str
    :param checksum_file_name: Name of the checksum file to be generated.
    :type checksum_file_name: str
    :param pgp_key_id: The ID of the PGP key used to sign the checksum file.
    :type pgp_key_id: str
    :return: None
    """
    __create_named_unsigned_checksum_file(file_name, checksum_file_name)
    sign_checksum_file(checksum_file_name, pgp_key_id)


def __create_named_unsigned_checksum_file(file_name: str, checksum_file_name: str):
    """
    Creates a named unsigned checksum file for the given file using its checksum value.

    This function computes the checksum of a file and then writes it to a specified checksum
    file.

    :param file_name: The name of the file for which the checksum is to be calculated.
    :type file_name: str
    :param checksum_file_name: The name of the file where the checksum should be stored.
    :type checksum_file_name: str
    :return: None
    """
    file_sum = __create_file_sum(file_name)
    write_checksum_file(file_name, file_sum, checksum_file_name)


def __create_file_sum(file_name: str) -> str:
    """
    Generates the SHA-256 hash of the given file's content.

    This function computes the SHA-256 checksum for the given file by reading it
    in binary mode and updating the hash digest iteratively for each block of 4096
    bytes. It returns the resulting SHA-256 hexdigest as a string.

    :param file_name: The name of the file for which the SHA-256 checksum is to
        be computed.
    :type file_name: str
    :return: The SHA-256 hexdigest of the file content.
    :rtype: str
    """
    sha256_hash = hashlib.sha256()
    with open(file_name, "rb") as f:
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)
        return sha256_hash.hexdigest()


def write_checksum_file(file_name_with_path: str, file_sum: str, checksum_file_name: str):
    """
    Writes a checksum file for a given file. The checksum file contains information
    about the file's name, its size, and the checksum value. This function
    obtains the file statistics using its path and writes the required data
    to a newly created checksum file.

    :param file_name_with_path: The path to the file for which the checksum file
        will be generated.
    :type file_name_with_path: str
    :param file_sum: The checksum value of the file that will be written into
        the checksum file.
    :type file_sum: str
    :param checksum_file_name: The name of the output checksum file that will
        be created.
    :type checksum_file_name: str
    :return: None.
    :rtype: None
    """
    file_stats: stat_result = os.stat(file_name_with_path)
    with open(checksum_file_name, 'w') as f:
        f.write(__generate_checksum_file_content(file_name_with_path, file_stats, file_sum))


def __generate_checksum_file_content(file_name_with_path: str, file_stats: stat_result, file_sum: str):
    """
    Generates a string representing the content of a checksum file based on the
    given file name, file statistics, and checksum value.

    This function utilizes a pre-defined template to format the file name, file
    size, and checksum into a readable string representation. It uses the SHA256
    hash algorithm identifier for the checksum in the output string.

    :param file_name_with_path: The full path string of the file including
        the file name which will be used for generating the checksum content.
        Should be a valid file path.
    :type file_name_with_path: str
    :param file_stats: A stat_result object containing the metadata of the file
        (e.g., size, modified time). This function specifically uses the `st_size`
        attribute from the file statistics for the file size in bytes.
    :type file_stats: stat_result
    :param file_sum: The SHA256 checksum string for the file, which is included
        in the generated checksum content.
    :type file_sum: str
    :return: A formatted string containing the file name, file size in bytes, and
        the checksum value following the predefined template.
    :rtype: str
    """
    template = string.Template('#$FILE_NAME $FILE_SIZE_BYTES bytes\nSHA256 ($FILE_NAME) = $FILE_SUM')
    return template.substitute({
        'FILE_NAME': os.path.basename(file_name_with_path),
        'FILE_SIZE_BYTES': file_stats.st_size,
        'FILE_SUM': file_sum
    })


def sign_checksum_file(checksum_file_name: str, pgp_key_id: str):
    """
    Signs a given checksum file using the specified PGP key ID.

    This function uses the gnupg GPG signing feature to sign
    a checksum file. The signature is appended to the file,
    ensuring its authenticity and integrity.

    :param checksum_file_name: The name of the checksum file to sign.
    :type checksum_file_name: str
    :param pgp_key_id: The PGP key ID to be used for signing the checksum file.
    :type pgp_key_id: str
    :return: None
    """
    gnupg.GPG().sign_file(checksum_file_name, keyid=pgp_key_id, output=checksum_file_name)
